/*
** Module   : InputOutputExtentParameter.java
** Abstract : Base class for special wrapper classes used for variables passed as 
**            input-output extent parameters to functions or procedures.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20140613 Created initial version.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.Array;

/**
 * Special wrapper used for variables passed as input-output extent parameters to 
 * functions or procedures. On normal function and procedure exit, if the reference 
 * for an output or input-output extent parameter has changed, this will update 
 * the caller's variable reference too.
 */
public abstract class InputOutputExtentParameter<T extends BaseDataType> 
extends AbstractExtentParameter<T>
{
   /**
    * The constructor responsible for instance initialization by calling 
    * {@link AbstractExtentParameter#AbstractExtentParameter(BaseDataType[])}.
    */
   public InputOutputExtentParameter()
   {
      super(null);
   }
   
   /**
    * This constructor allows to provide additional initialization logic of the variable 
    * reference in the extending classes. The initialization logic is encapsulated in 
    * {@link AbstractExtentParameter#AbstractExtentParameter(BaseDataType[])}.
    * 
    * @param   variable
    *          A valid and initialized variable reference.
    */
   protected InputOutputExtentParameter(T[] variable)
   {
      super(variable);
   }
   
   /**
    * This constructor allows to provide additional initialization logic of the variable 
    * reference in the extending classes. The initialization logic is encapsulated in 
    * {@link AbstractExtentParameter#AbstractExtentParameter(BaseDataType[])}.
    * 
    * @param   variable
    *          A valid and initialized variable reference.
    */
   @Override
   public T[] initParameter()
   {
      T[] var = getVariable();
      return initParameter(var.length);
   }

   /**
    * Initializes the indeterminate or determinate extent parameter reference depending
    * on the {@code extent} value. A zero value will initialize an indeterminate
    * extent, a positive integer will initialize a determinate extent. The extent
    * will be registered with {@link ArrayAssigner} accordingly. 
    * 
    * @param   extent
    *          The parameter extent size.
    * 
    * @return  The initialized indeterminate extent parameter reference.
    */
   @SuppressWarnings("unchecked")
   @Override
   public T[] initParameter(int extent)
   {
      T[] param = getParameter();
      if (param != null)
      {
         // the parameter is already initialized, just return it
         return param;
      }
      
      T[] var = getVariable();
      
      if (var.length > 0 && extent > 0 && var.length != extent)
      {
         character thisName = new character();
         if (SessionUtils._isRemote())
         {
            thisName.assign("(Remote client)");
         }
         else
         {
            handle pthis = ProcedureManager.thisProcedure();
            thisName.assign(ProcedureManager.getRelativeName(pthis.get()));
         }

         // not fixing the target and the extent sizes don't match
         String msg = "Extent parameter dimension of %d from procedure %s is mismatched with " +
               "sub-procedure %s %s parameter dimension of %d ...(328)";
         msg = String.format(msg,
                             extent,
                             thisName.toStringMessage(),
                             "TODO" /*caller.getInternalEntryName()*/,
                             "TODO" /*name*/,
                             var.length);
               
         ErrorManager.recordOrThrowError(11428, msg, false);
      }
      
      Class<?> varComponentClass = var.getClass().getComponentType();
      
      if (var.length < extent)
      {
         // The passed-in variable is an unfixed indeterminate extent, allocate
         // the parameter variable to the expected extent size. 
         // Don't use ArrayAssigner.resize in this case, as it also copies decimal
         // precision. According to Progress behavior the newly allocated array 
         // will have the default decimal precision.
         param = (T[]) Array.newInstance(varComponentClass, extent);
         for (int i = 0; i < extent; i++)
         {
            param[i] = (T) BaseDataType.generateUnknown(varComponentClass);
         }
         
         setParameter(param);
      }
      else
      {
         // The passed-in variable is a fixed extent, allocate the parameter 
         // variable to its extent size. 
         param = (T[]) Array.newInstance(varComponentClass, var.length);
         for (int i = 0; i < extent; i++)
         {
            // duplicate the array content
            param[i] = (T) var[i].duplicate();
            if (decimal.class.equals(varComponentClass))
            {
               // and set to the default decimal precision
               ((decimal) param[i]).setPrecision(decimal.MAX_SCALE);
            }
         }
         
         setParameter(param);
         
         if (extent == 0)
         {
            // the result is registered as dynamic array 
            ArrayAssigner.registerDynamicArray(param);
            
            // make sure this extent parameter is tracked by ArrayAssigner,
            // to handle parameter value updates on resizes for example
            ArrayAssigner.registerExtentParameter(this);
         }
      }
      
      return param;
   }
}
