/*
** Module   : OutputExtentParameter.java
** Abstract : Special wrapper used for variables passed as output extent parameters to
**            functions or procedures.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130209 Created initial version.
** 002 CA  20131013 Added no-op scopeDeleted() method, required by the changes in Scopeable 
**                  interface.
** 003 GES 20140122 Added safety code to avoid abends when silent errors occur.
** 004 GES 20140206 Added safety code to avoid abends when the persistence layer is not active. 
** 005 HC  20140413 Fixed extent parameter initialization.
** 006 HC  20140613 Improved extent parameter support.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.Array;

/**
 * Special wrapper used for variables passed as output extent parameters to functions 
 * or procedures. On normal function and procedure exit, if the reference for an output 
 * or input-output extent parameter has changed, this will update the caller's variable 
 * reference too.
 */
public abstract class OutputExtentParameter<T extends BaseDataType> 
extends AbstractExtentParameter<T>
{
   /**
    * The constructor responsible for instance initialization by calling 
    * {@link AbstractExtentParameter#AbstractExtentParameter(BaseDataType[])}.
    */
   public OutputExtentParameter()
   {
      super(null);
   }
   
   /**
    * This constructor allows to provide additional initialization logic of the variable 
    * reference in the extending classes. The initialization logic is encapsulated in 
    * {@link AbstractExtentParameter#AbstractExtentParameter(BaseDataType[])}.
    * 
    * @param   variable
    *          A valid and initialized variable reference.
    */
   protected OutputExtentParameter(T[] variable)
   {
      super(variable);
   }
   
   /**
    * Initializes the indeterminate extent parameter reference.
    *
    * @return  The initialized indeterminate extent parameter reference.
    */
   @Override
   public T[] initParameter()
   {
      T[] param = initParameter(0);
      return param;
   }

   /**
    * Initializes the indeterminate or determinate extent parameter reference depending
    * on the {@code extent} value. A zero value will initialize an indeterminate
    * extent, a positive integer will initialize a determinate extent. The extent
    * will be registered with {@link ArrayAssigner} accordingly. 
    * 
    * @param   extent
    *          The parameter extent size.
    * 
    * @return  The initialized indeterminate extent parameter reference.
    */
   @SuppressWarnings("unchecked")
   public T[] initParameter(int extent)
   {
      // get the parameter reference
      T[] param = getParameter();
      if (param != null)
      {
         // the parameter is already initialized, just return it
         return param;
      }
      
      // initialize the parameter reference according to the variable reference
      T[] var = getVariable();
      Class<?> varComponentClass = var.getClass().getComponentType();
      param = (T[]) Array.newInstance(varComponentClass, extent);
      for (int i = 0; i < extent; i++)
      {
         param[i] = (T) BaseDataType.generateDefault(varComponentClass); 
      }
      
      setParameter(param);
      
      // handle ArrayAssigner registrations for indeterminate extents
      if (extent == 0)
      {
         // indeterminate extent is registered as a dynamic array
         ArrayAssigner.registerDynamicArray(param);
         
         // make sure this extent parameter is tracked by ArrayAssigner,
         // to handle parameter value updates on resizes for example
         ArrayAssigner.registerExtentParameter(this);
      }
      
      return param;
   }

   /**
    * Output parameter validation logic. The method checks for runtime
    * compatibility of the extent variable reference and parameter reference
    * and raises an error if validation fails. 
    */
   @Override
   protected void validateAssignment()
   {
      T[] var = getVariable();
      T[] param = getParameter();
      
      // is the assignment fixing the target extent variable?
      if (var.length > 0 && var.length != param.length)
      {
         // no, raise an error
         character thisName = new character();
         if (SessionUtils._isRemote())
         {
            thisName.assign("(Remote client)");
         }
         else
         {
            handle pthis = ProcedureManager.thisProcedure();
            thisName.assign(ProcedureManager.getRelativeName(pthis.get()));
         }

         // not fixing the target and the extent sizes don't match
         String msg = "Extent parameter dimension of %d from procedure %s is mismatched with " +
               "sub-procedure %s %s parameter dimension of %d ...(328)";
         msg = String.format(msg,
                             var.length,
                             thisName.toStringMessage(),
                             "TODO" /*caller.getInternalEntryName()*/,
                             "TODO" /*name*/,
                             param.length);
               
         ErrorManager.recordOrThrowError(11428, msg, false);
      }
   }
}
